from granian.constants import Interfaces, Loops
from granian.log import LogLevels
from granian.server import Server


def main() -> None:
    server = Server(
        target="p_cast.app:create_app",
        factory=True,
        address="0.0.0.0",  # noqa: S104
        port=3000,
        interface=Interfaces.ASGI,
        loop=Loops.uvloop,
        log_level=LogLevels.debug,
    )
    server.serve()


if __name__ == "__main__":
    main()



import asyncio
import contextlib
import logging
import tempfile
import typing
from collections.abc import AsyncIterator
from functools import partial
from typing import override

from pychromecast import Chromecast
from pychromecast.controllers.media import MediaController
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route
from starlette.staticfiles import StaticFiles

from p_cast.cast import find_chromecast, get_local_ip, subscribe_to_stream
from p_cast.config import StreamConfig
from p_cast.device import SinkController
from p_cast.ffmpeg import create_ffmpeg_stream_command

logger = logging.getLogger(__name__)


class StaticFilesWithCORS(BaseHTTPMiddleware):
    @override
    async def dispatch(
        self,
        request: Request,
        call_next: RequestResponseEndpoint,
    ) -> Response:
        client = typing.cast("tuple[str, str]", request.scope["client"])
        method = typing.cast("str", request.scope["method"])
        path = typing.cast("str", request.scope["path"])

        response: Response = await call_next(request)

        content_length = int(response.headers["Content-Length"]) / 1024

        logger.debug(
            "[%s:%s] %s - %s - %s KiB",
            client[0],
            client[1],
            method,
            path,
            content_length,
        )
        response.headers["Access-Control-Allow-Origin"] = "*"
        response.headers["Access-Control-Allow-Credentials"] = "true"
        response.headers["Cache-Control"] = "no-cache"

        del response.headers["etag"]
        del response.headers["accept-ranges"]
        return response


@contextlib.asynccontextmanager
async def lifespan(
    app: Starlette,
    chromecast: Chromecast,
) -> AsyncIterator[None]:
    sink_controller = SinkController(chromecast=chromecast)
    await sink_controller.init()

    stream_config = StreamConfig(acodec="aac", bitrate="256k")

    app.state.media_controller = chromecast

    async def subscribe(mc: MediaController, local_ip: str) -> None:
        await asyncio.sleep(2)
        subscribe_to_stream(mc, local_ip, stream_config)

    with tempfile.TemporaryDirectory() as stream_dir:
        sink = await sink_controller.get_sink_name()
        logger.info("Casting from sink: %s", sink)
        ffmpeg_command = create_ffmpeg_stream_command(
            sink=f"{sink}.monitor",
            stream_dir=stream_dir,
            config=stream_config,
        )
        logger.info("Starting ffmpeg: %s", " ".join(ffmpeg_command))

        await asyncio.create_subprocess_exec(*ffmpeg_command)

        stream_app = StaticFilesWithCORS(StaticFiles(directory=stream_dir))

        app.mount("/stream", stream_app)

        local_ip = get_local_ip()

        task = asyncio.create_task(subscribe(chromecast.media_controller, local_ip))

        yield

        task.cancel()
        await sink_controller.close()


def get_media_controller(request: Request) -> MediaController:
    return typing.cast("MediaController", request.app.state.media_controller)  # pyright: ignore[reportAny]


async def pause(request: Request) -> Response:
    media_controller = get_media_controller(request)
    media_controller.pause()
    return Response(content="OK")


async def play(request: Request) -> Response:
    media_controller = get_media_controller(request)
    media_controller.play()
    media_controller.seek(None)  # type: ignore[arg-type] # pyright: ignore[reportArgumentType]
    return Response(content="OK")


def create_app() -> Starlette:
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger(__name__).setLevel(logging.INFO)
    # logger = logging.getLogger(__name__)
    chromecast = find_chromecast()

    middleware = [
        Middleware(
            GZipMiddleware,
            minimum_size=1,
            compresslevel=9,
        ),
        Middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_headers=["*"],
            allow_credentials=True,
            allow_methods=["*"],
        ),
    ]

    return Starlette(
        routes=[
            Route("/pause", endpoint=pause),
            Route("/play", endpoint=play),
        ],
        lifespan=partial(lifespan, chromecast=chromecast),
        middleware=middleware,
    )



from p_cast.config import StreamConfig


def create_ffmpeg_stream_command(
    sink: str,
    stream_dir: str,
    config: StreamConfig | None = None,
) -> list[str]:
    if config is None:
        config = StreamConfig()

    acodec = []
    if config.acodec == "aac":
        acodec = ["-c:a", "aac", "-profile:a", "aac_low"]
    return [
        "ffmpeg",
        "-y",
        "-loglevel",
        "quiet",
        "-f",
        "pulse",
        "-i",
        sink,
        ## BLACK SQUARE --
        # "-map",
        # "0:v",
        # "-map",
        # "1:a",
        # "-f",
        # "lavfi",
        # "-re",
        # "-i",
        # "color=size=2x2:rate=50:color=black",
        # "-c:v",
        # "libx264",
        # "-preset",
        # "ultrafast",
        # "-tune",
        # "zerolatency",
        # "-b:v",
        # "48k",
        # "-pix_fmt",
        # "yuv420p",
        # "-profile:v",
        # "baseline",
        # "-level",
        # "3.0",
        # "-g",
        # "25",
        # "-keyint_min",
        # "25",
        # "-sc_threshold",
        # "0",
        ## BLACK SQUARE --
        ## COPY --
        # "-c:v",
        # "copy",
        ## COPY --
        *acodec,
        "-b:a",
        config.bitrate,
        "-ar",
        str(config.sampling_frequency),
        "-fflags",
        "nobuffer",
        "-flags",
        "low_delay",
        "-flush_packets",
        "1",
        "-max_delay",
        "0",
        "-muxdelay",
        "0",
        "-strict",
        "experimental",
        "-avioflags",
        "direct",
        "-fflags",
        "discardcorrupt",
        "-probesize",
        "32",
        "-analyzeduration",
        "0",
        "-f",
        "hls",
        "-hls_time",
        "0.5",
        "-hls_list_size",
        "6",
        "-hls_delete_threshold",
        "1",
        "-hls_flags",
        "split_by_time+program_date_time+independent_segments+append_list",
        "-hls_segment_type",
        config.hls_segment_type,
        "-master_pl_name",
        "index.m3u8",
        "-var_stream_map",
        "a:0",
        stream_dir + "/main_stream.m3u8",
    ]



import asyncio
import typing

import pulsectl
import pulsectl_asyncio
from pychromecast import Chromecast

from p_cast.exceptions import SinkError


class SinkController:
    _sink_module_id: int
    _listener: asyncio.Task[None]

    def __init__(self, chromecast: Chromecast) -> None:
        self._cast = chromecast
        self._sink_module_id = -1

    async def init(self) -> None:
        self._pulse = pulsectl_asyncio.PulseAsync("p-cast-client")
        await self._pulse.connect()
        self._sink_module_id = await self.create_sink("P-Cast")
        self._listener = asyncio.create_task(self.subscribe())

    async def create_sink(self, name: str) -> int:
        sink_properties = [
            f"device.description={name}",
            "channelmix.min-volume=5.0",
            "channelmix.max-volume=5.0",
            "channelmix.normalize=true",
        ]
        module_args = [
            f"sink_name={name}",
            f'sink_properties="{" ".join(sink_properties)}"',
        ]

        module_id = await self._pulse.module_load(
            "module-null-sink",
            " ".join(module_args),
        )
        return typing.cast("int", module_id)

    async def get_sink(self) -> pulsectl.PulseSinkInfo:
        if self._sink_module_id == -1:
            msg = "Uninitialized device"
            raise SinkError(msg)
        return typing.cast(
            "pulsectl.PulseSinkInfo",
            await self._pulse.get_sink_by_name("P-Cast"),
        )

    async def get_sink_name(self) -> str:
        sink = await self.get_sink()
        return sink.name  # pyright: ignore[reportAttributeAccessIssue]

    async def close(self) -> None:
        await self._pulse.module_unload(self._sink_module_id)
        self._listener.cancel()

    def get_volume(self, sink: pulsectl.PulseSinkInfo) -> float:
        return typing.cast("int", sink.volume.values[0])

    def get_mute(self, sink: pulsectl.PulseSinkInfo) -> bool:
        return bool(sink.mute)  # pyright: ignore[reportAttributeAccessIssue]

    async def subscribe(self) -> None:
        sink = await self.get_sink()
        current_volume = self.get_volume(sink)
        current_mute = self.get_mute(sink)

        async for event in self._pulse.subscribe_events(
            pulsectl.PulseEventMaskEnum.sink,  # pyright: ignore[reportAttributeAccessIssue]
        ):
            if event.index != sink.index:  # pyright: ignore[reportAttributeAccessIssue]
                continue
            if event.t != pulsectl.PulseEventTypeEnum.change:  # pyright: ignore[reportAttributeAccessIssue]
                continue

            changed_sink = await self.get_sink()
            changed_volume = self.get_volume(changed_sink)
            changed_mute = self.get_mute(changed_sink)
            if changed_volume != current_volume:
                self._cast.set_volume(volume=changed_volume)
                current_volume = changed_volume
            if changed_mute != current_mute:
                self._cast.set_volume_muted(muted=changed_mute)
                current_mute = changed_mute



from dataclasses import dataclass
from typing import Literal

from p_cast.exceptions import StreamError

type AudioCodec = Literal["aac"]
type HlsSegmentType = Literal["mpegts", "fmp4"]

sample_rates: dict[AudioCodec, list[int]] = {
    "aac": [
        96000,
        88200,
        64000,
        48000,
        44100,
        32000,
        24000,
        22050,
        16000,
        12000,
        11025,
        8000,
        7350,
    ],
}


@dataclass
class StreamConfig:
    acodec: AudioCodec = "aac"
    bitrate: str = "192k"
    sampling_frequency: int = 48000
    hls_segment_type: HlsSegmentType = "mpegts"

    def __post_init__(self) -> None:
        if self.sampling_frequency not in sample_rates.get(self.acodec, []):
            msg = "Invalid sampling rate"
            raise StreamError(msg)

    @property
    def chromecast_hls_segment_type(self) -> str | None:
        match self.hls_segment_type:
            case "mpegts":
                return "ts_aac"
            case "fmp4":
                return "fmp4"



import subprocess


def get_default_sink() -> str:
    result = subprocess.run(  # noqa: S603
        ["pactl", "get-default-sink"],  # noqa: S607
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        text=True,
        check=True,
    )
    return result.stdout.strip()



import logging
import socket
import sys
import time
import typing
from uuid import UUID

import pychromecast
import zeroconf
from pychromecast.controllers.media import (
    STREAM_TYPE_LIVE,
    MediaController,
)
from pychromecast.discovery import CastBrowser, SimpleCastListener

from p_cast.config import StreamConfig

logger = logging.getLogger(__name__)


def find_chromecast() -> pychromecast.Chromecast:
    services: dict[UUID, str] = {}
    zconf = zeroconf.Zeroconf()

    def add_callback(device_id: UUID, service: str) -> None:
        logger.info("[%s] added: %s", device_id, service)
        services[device_id] = service

    browser = CastBrowser(
        SimpleCastListener(
            add_callback=add_callback,
        ),
        zconf,
    )
    browser.start_discovery()
    time.sleep(2)

    chromecasts, browser = pychromecast.get_listed_chromecasts(
        uuids=list(services.keys()),
    )
    if len(chromecasts) == 0:
        sys.exit(1)

    cast = chromecasts[0]
    cast.wait()
    pychromecast.discovery.stop_discovery(browser)
    return cast


def get_local_ip() -> str:
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(("8.8.8.8", 80))
        return typing.cast("str", s.getsockname()[0])
    finally:
        s.close()


def subscribe_to_stream(
    mc: MediaController,
    local_ip: str,
    config: StreamConfig,
) -> None:
    url = f"http://{local_ip}:3000/stream/index.m3u8"
    logger.info("Subscribing to: %s", url)
    mc.play_media(
        url,
        content_type="application/vnd.apple.mpegurl",
        title="p-cast stream",
        stream_type=STREAM_TYPE_LIVE,
        media_info={
            "hlsSegmentFormat": config.chromecast_hls_segment_type,
        },
    )
    mc.seek(999999999)
